package net.ccbluex.liquidbounce.features.module.modules.misc

import net.ccbluex.liquidbounce.event.tickHandler
import net.ccbluex.liquidbounce.features.module.Category
import net.ccbluex.liquidbounce.features.module.ClientModule
import net.ccbluex.liquidbounce.utils.aiming.Rotation
import net.ccbluex.liquidbounce.utils.aiming.SlowStart
import net.ccbluex.liquidbounce.utils.aiming.anglesmooth.*
import net.ccbluex.liquidbounce.utils.aiming.setRotation
import net.ccbluex.liquidbounce.utils.client.chat
import net.ccbluex.liquidbounce.utils.entity.rotation
import net.ccbluex.liquidbounce.utils.math.toBlockPos
import net.minecraft.util.math.BlockPos
import java.util.*
import kotlin.math.sqrt

object ModulePathFinder : ClientModule("PathFinder", Category.MISC) {
    //settings
    private var angleSmooth = choices(this, "AngleSmooth") {
        arrayOf(
            LinearAngleSmoothMode(it),
            SigmoidAngleSmoothMode(it),
            AccelerationSmoothMode(it),
            BezierAngleSmoothMode(it),
            ConditionalLinearAngleSmoothMode(it)
        )
    }

    private val slowStart = tree(SlowStart(this))

    private val target by text("Target", "0 0 0")

    // variables
    private val diamonds = ArrayList<BlockPos>()
    private var timer = 0

    var targetPath = LinkedList<BlockPos>()
    var currentTargetVec3d = BlockPos.ORIGIN
    var reached = false
    private var wait = true

    override fun enable() {
        reached = true
        timer = 0
        diamonds.clear()
        wait = true

        if (!findExamplePath()) {
            chat("Failed to find a path!")
        }
    }

    val repeatable = tickHandler {
        if (targetPath.isEmpty()) {
            enabled = false
            return@tickHandler
        }

        executePath()
    }

    // find a path to destination and return to where we are now. The path is in "p" variable. it will also get it ready for "executePath()" to start working
    // return success or not
    // This is a simple example of the path finding function.
    private fun findExamplePath(): Boolean {
        // get target
        val xyz = target.split(" ")
        val t = BlockPos(xyz[0].toInt(), xyz[1].toInt(), xyz[2].toInt())
        // find path
        val path = findPath(t)
        if (path.isEmpty()) return false
        // by doing this, simply adding a way back.
        path.addAll(path.reversed())

        // prepare
        targetPath = path
        currentTargetVec3d = targetPath.poll()

        // start executing path
        reached = false

        return true
    }

    // should be run every tick to execute path.
    // core method
    private fun executePath() {
        if (reached)
            return

        val finalTarget = currentTargetVec3d.toBottomCenterPos()
        val dx = finalTarget.x - player.pos.x
        val dz = finalTarget.z - player.pos.z
        val dis = sqrt(dx * dx + dz * dz)

        if (dis < 0.7) {
            if (targetPath.isEmpty()) {
                reached = true
                mc.options.forwardKey.isPressed = false
                return
            }

            currentTargetVec3d = targetPath.poll()
        } else {
            val rotation = Rotation.lookingAt(finalTarget.add(0.0, 1.0, 0.0), player.eyePos)
            val targetRotation = angleSmooth.activeChoice.limitAngleChange(
                slowStart.rotationFactor,
                player.rotation,
                rotation,
                null,
                null
            )
            player.setRotation(targetRotation)

            mc.options.forwardKey.isPressed = true

            if (player.horizontalCollision && player.isOnGround) {
                mc.options.forwardKey.isPressed = false
                player.jump()
            }
        }
    }

    override fun disable() {
        mc.options.forwardKey.isPressed = false
    }

    // find a path to tartget bockpos.
    // core method
    private fun findPath(target: BlockPos): LinkedList<BlockPos> {
        val cur = player.pos.toBlockPos()

        val visited = HashSet<BlockPos>()
        val parent = HashMap<BlockPos, BlockPos>()
        val queue = LinkedList<BlockPos>()

        queue.add(cur)
        visited.add(cur)

        val t = target.add(0, 1, 0)
        while (!queue.isEmpty()) {
            if (visited.size >= 500) {
                chat("error!!")
                return LinkedList()
            }

            val a = queue.poll()

            if (a.equals(t)) {
                val path = LinkedList<BlockPos>()
                var b = a
                path.add(b)
                while (parent.containsKey(b)) {
                    val p = parent[b]
                    path.add(p!!)
                    b = p
                }
                path.reverse()
                return path
            }

            for (l in offset) {
                val new = BlockPos(a.x + l[0], a.y + l[1], a.z + l[2])
                if (!world.getBlockState(new).isAir || world.getBlockState(new.down()).isAir || !world.getBlockState(new.up()).isAir)
                    continue

                if (visited.contains(new))
                    continue

                visited.add(new)
                queue.add(new)
                parent[new] = a
            }
        }

        return LinkedList()
    }

    // renders below

    /*
    @Suppress("unused")
    val renderHandler = handler<WorldRenderEvent> { event ->
        val matrixStack = event.matrixStack

        if (p.isNotEmpty()) drawBoxMode(matrixStack)
    }

    private fun drawBoxMode(matrixStack: MatrixStack): Boolean {
        var dirty = false

        renderEnvironmentForWorld(matrixStack) { dirty = drawInternal(p) }

        return dirty
    }

    private fun WorldRenderEnvironment.drawInternal(
        blocks: LinkedList<BlockPos>
    ): Boolean {
        var dirty = false

        BoxRenderer.drawWith(this) {
            for (blockPos in blocks) {
                val blockState = blockPos.getState() ?: continue

                val outlineShape = blockState.getOutlineShape(world, blockPos)
                val boundingBox = if (outlineShape.isEmpty) {
                    FULL_BOX
                } else {
                    outlineShape.boundingBox
                }

                withPositionRelativeToCamera(blockPos.toVec3d()) {
                    drawBox(
                        boundingBox,
                        faceColor = Color4b.GREEN,
                        outlineColor = Color4b.TRANSPARENT
                    )
                }

                dirty = true
            }
        }

        return dirty
    }
    */

    // using a offset instead of foreach block is to prioritize specific position offset like straight instead of moving in x and z at the same time which is dangerous currently.
    val offset = arrayOf(
        arrayOf(-1, -1, 0),
        arrayOf(-1, 0, 0),
        arrayOf(-1, 1, 0),
        arrayOf(0, -1, -1),
        //arrayOf(0, -1, 0),
        arrayOf(0, -1, 1),
        arrayOf(0, 0, -1),
        arrayOf(0, 0, 1),
        //arrayOf(0, 1, 0),
        //arrayOf(0, 0, 0),
        arrayOf(0, 1, -1),
        arrayOf(0, 1, 1),
        arrayOf(1, -1, 0),
        arrayOf(1, 0, 0),
        arrayOf(1, 1, 0),

        // we better not move in following direction
        arrayOf(-1, -1, -1),
        arrayOf(-1, -1, 1),
        arrayOf(-1, 0, -1),
        arrayOf(-1, 0, 1),
        arrayOf(-1, 1, -1),
        arrayOf(-1, 1, 1),
        arrayOf(1, -1, -1),
        arrayOf(1, -1, 1),
        arrayOf(1, 0, -1),
        arrayOf(1, 0, 1),
        arrayOf(1, 1, -1),
        arrayOf(1, 1, 1)
    )
}
